/*

  P-Code interpreter (to run the apple pascal system)
  Copyright (C) 2000 Mario Klebsch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <curses.h>
#include <term.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "psystem.h"
#include "Memory.h"
#include "Term.h"

static int Term=-1;

/*

  erase line=0
  erase to end of screen=11
  erase screen=12
  insert line=22
  move cursor home=25
  move cursor right=28
  erase to end of line=29
  gotoxy=30
  move cursor up=31

 */

void TermWrite(char ch, word Mode)
{
  static int state=0;
  static int x;
  int	     y;
  char	     *s;

  switch(state)
    {
    case 0:
      switch(ch)
	{
#ifdef XXX
	case 0:				/* erase line			    */
	  break;
#endif
	case 11:			/* erase to end of screen	    */
	  write(Term, clr_eos, strlen(clr_eos));
	  break;
	case 12:			/* erase screen			    */
	  write(Term, clear_screen, strlen(clear_screen));
	  break;
	case 13:			/* carriage return		    */
	  write(Term,"\r\n", 2);
	  break;
	case 16:			/* DLE prefix			    */
	  if (!(Mode&0x8))
	    state=3;
	  break;
	case 22:
	  write(Term, insert_line, strlen(insert_line));
	  break;
	case 25:			/* move cursor home		    */
	  write(Term, cursor_home, strlen(cursor_home));
	  break;
	case 28:			/* move cursor right		    */
	  write(Term, cursor_right, strlen(cursor_right));
	  break;
	case 29:			/* erase to end of line		    */
	  write(Term, clr_eol, strlen(clr_eol));
	  break;
	case 30:			/* gotoxy			    */
	  state=1;
	  break;
	case 31:			/* move cursor up		    */
	  write(Term, cursor_up, strlen(cursor_up));
	  break;
	default:
	  write(Term, &ch, 1);
	  break;
	}
      break;
    case 1:
      x=ch-' ';
      state=2;
      break;
    case 2:				/* gotoxy complete		    */
      y=ch-' ';
      s=tgoto(cursor_address, x, y);
      write(Term, s, strlen(s));
      state=0;
      break;
    case 3:				/* char after DLE		    */
      ch -= 0x20;
      while (ch--)
	write(Term, " ", 1);
      state=0;
      break;
    }
}

#define NUMBER(a) (sizeof(a)/sizeof(a[0]))
static struct translator
{
  byte	Key;
  char	*Pattern;
} *Keys=NULL;
static int NumKeys=0;

char TermRead(void)
{
  unsigned char	ch;

  static int	state=0;
  static int	BufIdx=0;
  static char	Buffer[10];

  if (state==0)
    while (1)
      {
	fd_set		fds;
	struct timeval	tv;
	FD_ZERO(&fds);
	FD_SET(Term, &fds);
	tv.tv_sec = 0;
	tv.tv_usec = 25000;
	
	if (select(Term+1, &fds, NULL, NULL, BufIdx?&tv:NULL)>0)
	  {				/* Ein Zeichen ist verfgbar	    */
	    struct translator *t;
	    int	Match=0;

	    read(Term, &ch, 1);		/* Lesen und in Puffer speichern    */

	    if (ch=='\0')	/* XXX Quick-Hack :-( */
	      continue;

	    Buffer[BufIdx++]=ch;
	    for (t=Keys; t< &Keys[NumKeys]; t++)
	      {
		int	i;
		for (i=0; i< BufIdx; i++)
		  if (Buffer[i] != t->Pattern[i])
		    goto next;
		if (!(t->Pattern[BufIdx]))	/* full Pattern match */
		  {
		    BufIdx=0;
		    return (t->Key);
		  }
		Match=1;
		break;
	      next:
	      }
	    if (!Match)
	      goto ret;
	  }
	else				/* Timeout			    */
	  {
	  ret:
	    if (BufIdx>1)
	      state=1;
	    else
	      BufIdx=0;
	    return(Buffer[0]);
	  }
      }
  else
    {
      ch=Buffer[state++];
      if (state >=BufIdx)
	state=0;
      return(ch);
    }
}

void TermStat(word Addr, Integer Offset, word Dummy)
{
  fd_set		fds;
  struct timeval	tv;
  FD_ZERO(&fds);
  FD_SET(Term, &fds);
  tv.tv_sec = 0;
  tv.tv_usec = 0;
	
  if (select(Term+1, &fds, NULL, NULL, &tv)>0)
    MemWr(Addr, 1);
  else
    MemWr(Addr, 0);
}

void TermOpen(void)
{
  int			s;
  struct sockaddr_in	addr;
  int			i;
  char			Buffer[256];

  unsigned char		TelnetInit[]=
    { 255, 253, 3,
      255, 251, 3,
      255, 251, 1
    };

  if ((s=socket(PF_INET, SOCK_STREAM, 0))<0)
    {
      perror("socket()");
      exit(1);
    }
  addr.sin_family=AF_INET;
  addr.sin_addr.s_addr=htonl(INADDR_LOOPBACK);
  addr.sin_port=htons(0);
  if (bind(s, (struct sockaddr*)&addr, sizeof(addr))<0)
    {
      perror("bind()");
      exit(1);
    }
  if (listen(s,1)<0)
    {
      perror("listen()");
      exit(1);
    }
  i=sizeof(addr);
  if (getsockname(s, (struct sockaddr*)&addr, &i)<0)
    {
      perror("getsockname()");
      exit(1);
    }

  snprintf(Buffer, sizeof(Buffer), "xterm -geometry 80x24 -e "
	   "telnet -E localhost %d&", ntohs(addr.sin_port));
  system(Buffer);

  i=sizeof(addr);
  if ((Term=accept(s, (struct sockaddr*)&addr, &i))<0)
    {
      perror("accept()");
      exit(1);
    }

  write(Term, TelnetInit, sizeof(TelnetInit));
  for (i=0; i<sizeof(TelnetInit); i++)
    TermRead();

  close(s);
  setterm("xterm");
  setupterm("xterm", Term, &i);

  NumKeys=4;
  Keys=(struct translator*)malloc(NumKeys*sizeof(Keys));

  Keys[0].Key=0x0b; Keys[0].Pattern="\033[A";
  Keys[1].Key=0x0a; Keys[1].Pattern="\033[B";
  Keys[2].Key=0x15; Keys[2].Pattern="\033[C";
  Keys[3].Key=0x08; Keys[3].Pattern="\033[D";
}

